#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <err.h>
#include <time.h>	/* clock_gettime */
#include <sys/time.h>	/* timespecsub */

#include <hammer/hammer.h>

int
match(const HParser *parser, const char *input, size_t sz)
{
	HParseResult *res;

	res = h_parse(parser, input, sz);
	if (res == NULL)
		return -1;
	sz = res->bit_length / 8;
	h_parse_result_free(res);

	return sz;
}

int
benchmark(const HParser *parser)
{
	static const char *inputs[] = {
	    "test@example.com",
	    "hans.wurst+kaese@gmail.com",
	    "!#$%&@valid.email",
	    "...@invalid",
	    "total@falsch.",
	    "auch@.falsch",
	    "genauso.@falsch.ey",
	    "bis@hier.richtig ABER",
	    "mett@wurst",
	    "hallo@",
	    "@dingens",
	    "@",
	    "",
	    NULL };
	
	static const long N = 5000;		/* runs per input */

	struct timespec t, t0, td;
	const char **p;
	int i, n;
	size_t sz;

	/* warmup */
	sz = strlen(inputs[0]);
	for (i = 0; i < N; i++)
		match(parser, inputs[0], sz);

	for (p = inputs; *p != NULL; p++) {
		printf("%30.30s  ", *p);
		fflush(stdout);

		sz = strlen(*p);
		clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &t0);
		for (i = 0; i < N; i++)
			n = match(parser, *p, sz);
		clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &t);
		timespecsub(&t, &t0, &td);

		if (n == sz)
			printf("match     ");
		else if (n < 0)
			printf("no match  ");
		else
			printf("partial   ");

		printf("%9.0f ns\n", (td.tv_sec * 1e9 + td.tv_nsec) / N);
	}

	return 0;
}

#include <getopt.h>

#define OTHER "!#$%&'*+/=?^_`{|}~-"

int
main(int argc, char **argv)
{
	int ch, x;

	/* character classes */
	HParser *lower = h_ch_range('a', 'z');
	HParser *upper = h_ch_range('Z', 'Z');
	HParser *digit = h_ch_range('0', '9');
	HParser *alpha = h_choice(lower, upper, NULL);
	HParser *other = h_in(OTHER, sizeof OTHER - 1);

	/* email addresses. cf. rfc 5322, section 3.4 */
	HParser *atext = h_choice(alpha, digit, other, NULL);
	HParser *word = h_many1(atext);
	HParser *dword = h_sequence(h_ch('.'), word, NULL);
	HParser *dotted = h_sequence(word, h_many(dword), NULL);
	HParser *address = h_sequence(dotted, h_ch('@'), dotted, NULL);

	HParser *parser = address;

	x = h_compile(parser, PB_REGULAR, NULL);
	if (x != 0) {
		fprintf(stderr, "h_compile: %d, ", x);
		switch (x) {
		case -1:
			fprintf(stderr, "parser not regular\n");
			break;
		case -2:
			fprintf(stderr, "invalid parameters\n");
			break;
		default:
			fprintf(stderr, "internal error\n");
		return 1;
		}
	}

	while ((ch = getopt(argc, argv, "b")) != -1) {
		switch (ch) {
		case 'b':
			return benchmark(parser);
		default:
			printf("usage: hammer [-b] string ...\n");
			return 1;
		}
	}

	for (int i = 1; i < argc; i++) {
		char *inp;
		size_t sz;
		int n;

		inp = argv[i];
		sz = strlen(inp);
		n = match(parser, inp, sz);
		if (n == -1)
			printf("no ");
		else if (n < sz)
			printf("partial ");
		printf("match: %.*s\n", n, inp);
	}
	
	return 0;
}
